package net.sf.doolin.bus;

import net.sf.doolin.bus.support.DefaultBus;

import java.util.concurrent.atomic.AtomicReference;

/**
 * Bus holder.
 *
 * @author Damien Coraboeuf
 */
public abstract class Bus {

    /**
     * Default bus
     */
    private static final AtomicReference<IBus> DEFAULT_BUS = new AtomicReference<IBus>(new DefaultBus());

    /**
     * Contextual bus
     */
    private static final ThreadLocal<IBus> CONTEXTUAL_BUS = new InheritableThreadLocal<IBus>();

    /**
     * Gets the current bus. First takes it from the thread context, then uses the default one.
     *
     * @return Bus to use in this context
     */
    public static IBus get() {
        // First in the context
        IBus bus = CONTEXTUAL_BUS.get();
        // ... then gets the default
        if (bus == null) {
            bus = DEFAULT_BUS.get();
        }
        // If none is defined, raises an error
        if (bus == null) {
            throw new IllegalStateException("No default bus nor a contextual one has been defined");
        } else {
            return bus;
        }
    }

    /**
     * Sets the bus in the thread context and returns the previous one.
     *
     * @param bus Bus to set (can be null)
     * @return Previous bus (can be null)
     */
    public static IBus setContextualBus(IBus bus) {
        IBus previous = CONTEXTUAL_BUS.get();
        CONTEXTUAL_BUS.set(bus);
        return previous;
    }


    /**
     * Sets the default bus to use and returns the previous one.
     *
     * @param bus Bus to set (can be null)
     * @return Previous bus (can be null)
     */
    public static IBus setDefaultBus(IBus bus) {
        return DEFAULT_BUS.getAndSet(bus);
    }

}
