package net.sf.doolin.bus.bean;

import net.sf.doolin.bus.Bus;
import net.sf.doolin.bus.support.PropertyChangeMessage;

/**
 * This class defines an object that can notify the changes of an object to the
 * bus.
 * <p/>
 * Notification of changes can be done on the spot, or buffered or completely
 * stopped.
 * <p/>
 * Any {@link BeanNotificationSupport} can have a parent and we can
 * therefore define a hierarchy of {@linkplain BeanNotificationSupport beans}.
 * Each level in the hierarchy participates in the notification of changes.
 * <p/>
 * For example, in the following hierarchy:
 * <p/>
 * <pre>
 * A {
 * 	B b;
 * }
 *
 * B {
 * 	C c;
 * }
 *
 * C {
 * 	String name;
 * }
 * </pre>
 * <p/>
 * If C notifies a change on <code>name</code> as in:
 * <p/>
 * <pre>
 * C {
 * 	...
 * 	notify(&quot;name&quot;, &quot;Old value&quot;, &quot;A new name&quot;);
 * 	...
 * }
 * </pre>
 * <p/>
 * Following events are generated on the bus:
 * <p/>
 * <pre>
 * C.name := &quot;A new name&quot;;
 * B.c.name := &quot;A new name&quot;;
 * A.b.c.name := &quot;A new name&quot;;
 * </pre>
 *
 * @author Damien Coraboeuf
 */
public class BeanNotificationSupport {

    // private transient ParentBean parentBean;
    private final Object holder;

    public BeanNotificationSupport(Object holder) {
        this.holder = holder;
    }

    // TODO Adds the Javadoc
    protected PropertyChangeMessage createMessage(String propertyName,
                                                  Object oldValue, Object newValue) {
        PropertyChangeMessage pcm = new PropertyChangeMessage(this.holder,
                propertyName, oldValue, newValue);
        return pcm;
    }

    /**
     * Notifies to the bus that a property has changed. If the notification mode
     * is BUFFERED, the change will be actually sent when notification mode is
     * restored to NORMAL.
     *
     * @param propertyName Name of the property
     * @param oldValue     Old value of the property
     * @param newValue     New value of the property
     * @see PropertyChangeMessage
     */
    public void notify(String propertyName, Object oldValue, Object newValue) {
        PropertyChangeMessage pcm = createMessage(propertyName,
                oldValue, newValue);
        Bus.get().publish(pcm);
        // Notification to the parent
        if (this.holder instanceof IBean) {
            IBean bean = (IBean) this.holder;
            // Parent change for the children
            if (oldValue instanceof IBean) {
                ((IBean) oldValue).setParentBean(null);
            }
            if (newValue instanceof IBean) {
                ((IBean) newValue).setParentBean(new ParentBean(
                        this.holder, propertyName));
            }
            // Parent notification
            IParentBean parentBean = bean.getParentBean();
            if (parentBean != null) {
                parentBean.notifyForChild(propertyName, oldValue,
                        newValue);
            }
        }
    }

}
