package net.sf.doolin.bus.support;

import net.sf.doolin.bus.BusLog;
import net.sf.doolin.bus.IBus;
import net.sf.doolin.bus.IBusInfo;
import net.sf.doolin.bus.Subscriber;
import org.apache.commons.lang.Validate;

import java.util.*;

public class DefaultBus implements IBus {

    private final LinkedList<Object> queue = new LinkedList<Object>();
    private final Map<Class<?>, Subscriptions<?>> subscriberMap = Collections
            .synchronizedMap(new HashMap<Class<?>, Subscriptions<?>>());

    private BusLog log = new DefaultBusLog();

    @Override
    public synchronized BusLog getLog() {
        return log;
    }

    @Override
    public synchronized void setLog(BusLog log) {
        this.log = log;
    }

    /**
     * Pushes the message on the queue.
     *
     * @param message Message which is published.
     */
    @Override
    public void publish(Object message) {
        Validate.notNull(message, "Message must not be null");
        boolean empty;
        synchronized (queue) {
            empty = queue.isEmpty();
        }
        if (empty) {
            process(message);
        } else {
            synchronized (queue) {
                queue.addLast(message);
            }
            processContinue();
        }
    }

    /**
     * Continue the processing
     */
    protected void processContinue() {
        Object message = null;
        synchronized (queue) {
            if (!queue.isEmpty()) {
                message = queue.removeFirst();
            }
        }
        if (message != null) {
            process(message);
        }
    }

    @Override
    public <M> void subscribe(Subscriber<M> subscriber) {
        Validate.notNull(subscriber, "Subscriber must not be null");
        // Gets the message class
        Class<M> messageClass = subscriber.getMessageClass();
        Validate.notNull(messageClass, "The Subscriber class must not be null");
        // Registers the subscriber for all categories
        Subscriptions<M> subscriptions;
        synchronized (subscriberMap) {
            @SuppressWarnings("unchecked")
            Subscriptions<M> s = (Subscriptions<M>) subscriberMap
                    .get(messageClass);
            subscriptions = s;
            if (subscriptions == null) {
                subscriptions = new Subscriptions<M>();
                subscriberMap.put(messageClass, subscriptions);
            }
        }
        // Add the subscriber
        subscriptions.add(subscriber);
        // Logs
        if (log != null) {
            log.registrationOk(subscriber);
        }
    }

    @Override
    public <M> void unsubscribe(Subscriber<M> subscriber) {
        Validate.notNull(subscriber, "Subscriber must not be null");
        // Gets the message class
        Class<M> messageClass = subscriber.getMessageClass();
        // Unregisters the subscriber for all categories
        synchronized (subscriberMap) {
            @SuppressWarnings("unchecked")
            Subscriptions<M> subscriptions = (Subscriptions<M>) subscriberMap
                    .get(messageClass);
            if (subscriptions != null) {
                subscriptions.remove(subscriber);
                if (log != null) {
                    log.unregistration(subscriber);
                }
            }
        }
    }

    /**
     * Processes one message
     *
     * @param message Message to process
     */
    protected <M> void process(M message) {
        // Traces the processing of the message
        messageReceived(message);
        // Messages are processed according to their class
        @SuppressWarnings("unchecked")
        Class<M> messageClass = (Class<M>) message.getClass();
        // Gets the list of subscribers for the message category
        Subscriptions<M> subscriptions = getSubscriptions(messageClass);
        if (subscriptions != null) {
            Collection<Subscription<M>> subscriptionList = subscriptions.getSubscriptions();
            // If there are some messages, process them
            if (!subscriptionList.isEmpty()) {
                int count = 0;
                boolean ok = true;
                Iterator<? extends Subscription<M>> i = subscriptionList.iterator();
                while (ok && i.hasNext()) {
                    Subscriber<M> subscriber = i.next().getSubscriber();
                    // Checks the subscription validity
                    if (subscriber.isValid()) {
                        if (subscriber.accept(message)) {
                            try {
                                count++;
                                // Logs the acceptance of the message
                                messageAccepted(message, subscriber);
                                subscriber.receive(message);
                            } catch (Exception ex) {
                                // Logs error
                                messageWithError(message, subscriber, ex);
                                ok = false;
                            }
                        }
                    } else {
                        // Removes the subscriber from the list
                        subscriptions.remove(subscriber);
                        if (log != null) {
                            log.unregistration(subscriber);
                        }
                    }
                }
                // No subscriber accepted the message
            } else {
                // Warning: no subscription
                messageNoSubscription(message);
            }
        } else {
            // Warning: no subscription
            messageNoSubscription(message);
        }
    }

    private void messageNoSubscription(Object message) {
        if (log != null) {
            log.messageNoSubscription(message);
        }
    }

    private <M> void messageWithError(M message, Subscriber<M> subscriber, Exception ex) {
        if (log != null) {
            log.messageWithError(message, subscriber, ex);
        }
    }

    protected <M> void messageAccepted(M message, Subscriber<M> subscriber) {
        if (log != null) {
            log.messageAccepted(message, subscriber);
        }
    }

    protected void messageReceived(Object message) {
        if (log != null) {
            log.messageReceived(message);
        }
    }

    @SuppressWarnings("unchecked")
    private <M> Subscriptions<M> getSubscriptions(Class<M> messageClass) {
        return (Subscriptions<M>) subscriberMap.get(messageClass);
    }

    @Override
    public void clear() {
        synchronized (subscriberMap) {
            for (Subscriptions<?> subscriptions : subscriberMap.values()) {
                subscriptions.clear();
            }
        }
    }

    @Override
    public IBusInfo getBusInfo() {
        BusInfo info = new BusInfo();
        // Number of pending messages
        int pendingMessages = queue.size();
        info.setPendingMessages(pendingMessages);
        // Collecting information about the subscribers
        synchronized (subscriberMap) {
            for (Map.Entry<Class<?>, Subscriptions<?>> entry : subscriberMap.entrySet()) {
                Class<?> messageClass = entry.getKey();
                Subscriptions<?> subscriptions = entry.getValue();
                SubscriptionClassInfo subscriptionClassInfo = new SubscriptionClassInfo(messageClass);
                List<? extends Subscription<?>> subscriptionList = subscriptions.getSubscriptions();
                for (Subscription<?> subscription : subscriptionList) {
                    subscriptionClassInfo.addSubscriptionInfo(subscription);
                }
                info.addSubscriptionClassInfo(subscriptionClassInfo);
            }
        }
        // OK
        return info;
    }

}
