package net.sf.doolin.bus.support;

import net.sf.doolin.bus.Bus;
import net.sf.doolin.util.Utils;
import net.sf.sido.common.Identifiable;

/**
 * This class defines a message that can be sent on the {@link Bus} to define a
 * property change in a bean.
 *
 * @author Damien Coraboeuf
 */
public class PropertyChangeMessage {

    private final Object bean;
    private final String propertyName;
    private final Object propertyValue;
    private final Object oldPropertyValue;

    /**
     * Constructor.
     *
     * @param bean         Bean whose property has changed
     * @param propertyName Path to the property
     * @param oldValue     Old property value
     * @param newValue     New value for the property
     */
    public PropertyChangeMessage(Object bean, String propertyName,
                                 Object oldValue, Object newValue) {
        this.bean = bean;
        this.propertyName = propertyName;
        this.oldPropertyValue = oldValue;
        this.propertyValue = newValue;
    }

    /**
     * Cancels the update
     */
    public void cancel() {
        Utils.setProperty(this.bean, this.propertyName, this.oldPropertyValue);
        Bus.get().publish(
                new PropertyChangeMessage(this.bean, this.propertyName,
                        this.propertyValue, this.oldPropertyValue));
    }

    /**
     * Returns the bean
     *
     * @return bean whose property has changed
     */
    public Object getBean() {
        return this.bean;
    }

    /**
     * Returns the previous value of the property
     *
     * @return Old property value
     */
    public Object getOldPropertyValue() {
        return this.oldPropertyValue;
    }

    /**
     * Returns the path to the changed property
     *
     * @return Property path
     */
    public String getPropertyName() {
        return this.propertyName;
    }

    /**
     * Returns the new value for the property
     *
     * @return Value
     */
    public Object getPropertyValue() {
        return this.propertyValue;
    }

    @Override
    public String toString() {
        String name = this.bean != null ? this.bean.getClass().getSimpleName()
                : "null";
        return String.format("[%s.%s=%s]", name, this.propertyName,
                this.propertyValue);
    }

    public boolean sameOrigin(Object source) {
        if (bean == source) {
            return true;
        } else if (bean instanceof Identifiable) {
            return ((Identifiable) bean).isSameAs(source);
        } else if (source instanceof Identifiable) {
            return ((Identifiable) source).isSameAs(bean);
        } else {
            return false;
        }
    }
}
