package net.sf.doolin.bus.support;

import net.sf.doolin.bus.Bus;
import net.sf.doolin.bus.Subscriber;
import net.sf.doolin.bus.SubscriberValidator;
import net.sf.doolin.util.Adapter;
import net.sf.doolin.util.Utils;

/**
 * Utility class to support property changes based on the {@link Bus}.
 *
 * @author Damien Coraboeuf
 * @see PropertyChangeSubscriber
 * @see PropertyChangeMessage
 */
public class PropertyChangeSupport {

    /**
     * Notifies the {@link Bus} that a property has changed
     *
     * @param bean     Bean whose property has changed
     * @param property Path to the changed property
     * @param oldValue New value for the property
     * @param newValue New value for the property
     */
    public static void change(Object bean, String property, Object oldValue,
                              Object newValue) {
        Bus.get().publish(
                new PropertyChangeMessage(bean, property, oldValue, newValue));
    }

    /**
     * Operates a two-ways connection between two beans and two properties. The
     * target property is immediately updated from the source.
     *
     * @param validator          Subscription validator
     * @param sourceBean         Source bean
     * @param sourcePropertyPath Path to the source property
     * @param targetBean         Target bean
     * @param targetPropertyPath Path to the target property
     * @see BothWayPropertyChangeSubscriber
     * @see #update(Object, String, Object, String, Adapter)
     */
    public static void connectBothWays(final SubscriberValidator validator,
                                       final Object sourceBean, final String sourcePropertyPath,
                                       final Object targetBean, final String targetPropertyPath) {
        Bus.get().subscribe(
                new BothWayPropertyChangeSubscriber(validator, sourceBean,
                        sourcePropertyPath, targetBean, targetPropertyPath));
        update(sourceBean, sourcePropertyPath, targetBean, targetPropertyPath,
                null);
    }

    /**
     * Connects a source property to a target property, in one way only.
     *
     * @param validator          Subscription validator
     * @param sourceBean         Source bean
     * @param sourcePropertyPath Path to the source property
     * @param targetBean         Target bean
     * @param targetPropertyPath Path to the target property
     */
    public static void connectOneWay(SubscriberValidator validator,
                                     Object sourceBean, String sourcePropertyPath,
                                     final Object targetBean, final String targetPropertyPath) {
        connectOneWay(validator, sourceBean, sourcePropertyPath, targetBean,
                targetPropertyPath, null);
    }

    /**
     * Connects a source property to a target property, in one way only, using
     * an adapter between the two properties.
     *
     * @param validator          Subscription validator
     * @param sourceBean         Source bean
     * @param sourcePropertyPath Path to the source property
     * @param targetBean         Target bean
     * @param targetPropertyPath Path to the target property
     * @param adapter            Adapter from the source property to the target property
     * @param <S>                Type of the source property
     * @param <T>                Type of the target property
     */
    public static <S, T> void connectOneWay(SubscriberValidator validator,
                                            Object sourceBean, String sourcePropertyPath,
                                            final Object targetBean, final String targetPropertyPath,
                                            final Adapter<S, T> adapter) {
        subscribe(new PropertyChangeSubscriber(getExecutionDescription(
                targetBean, targetPropertyPath), validator, sourceBean,
                sourcePropertyPath, new PropertyChangeListener() {
            @Override
            public void onChange(PropertyChangeMessage message) {
                Object propertyValue = message.getPropertyValue();
                if (adapter != null) {
                    @SuppressWarnings("unchecked")
                    S s = (S) propertyValue;
                    propertyValue = adapter.convertSubjectToTarget(s);
                }
                Utils.setProperty(targetBean, targetPropertyPath,
                        propertyValue);
            }
        }));
    }

    /**
     * Connects a source property to a target property, in one way only, and
     * updates the target immediately.
     *
     * @param validator          Subscription validator
     * @param sourceBean         Source bean
     * @param sourcePropertyPath Path to the source property
     * @param targetBean         Target bean
     * @param targetPropertyPath Path to the target property
     * @see #update(Object, String, Object, String, Adapter)
     */
    public static void connectOneWayAndUpdate(SubscriberValidator validator,
                                              Object sourceBean, String sourcePropertyPath,
                                              final Object targetBean, final String targetPropertyPath) {
        connectOneWayAndUpdate(validator, sourceBean, sourcePropertyPath,
                targetBean, targetPropertyPath, null);
    }

    /**
     * Connects a source property to a target property, in one way only, using
     * an adapter between the two properties, and updates the target property
     * immediately.
     *
     * @param validator          Subscription validator
     * @param sourceBean         Source bean
     * @param sourcePropertyPath Path to the source property
     * @param targetBean         Target bean
     * @param targetPropertyPath Path to the target property
     * @param adapter            Adapter from the source property to the target property
     * @param <S>                Type of the source property
     * @param <T>                Type of the target property
     * @see #update(Object, String, Object, String, Adapter)
     */
    public static <S, T> void connectOneWayAndUpdate(
            SubscriberValidator validator, Object sourceBean,
            String sourcePropertyPath, final Object targetBean,
            final String targetPropertyPath, Adapter<S, T> adapter) {
        connectOneWay(validator, sourceBean, sourcePropertyPath, targetBean,
                targetPropertyPath, adapter);
        update(sourceBean, sourcePropertyPath, targetBean, targetPropertyPath,
                adapter);
    }

    public static String getExecutionDescription(Object targetBean,
                                                 String targetPropertyPath) {
        return String.format("Sets property %s on %s", targetPropertyPath,
                Utils.simpleToString(targetBean, "<null>"));
    }

    /**
     * Utility method to subscribe a {@link Subscriber} to the {@link Bus}.
     *
     * @param subscriber Subscriber to subscribe
     */
    public static void subscribe(PropertyChangeSubscriber subscriber) {
        Bus.get().subscribe(subscriber);
    }

    /**
     * Registers a property subscriber that runs a task whenever the given
     * property is changed.
     *
     * @param validator    Subscription validator
     * @param bean         Bean whose property may be changed
     * @param propertyPath Path to the property
     * @param runnable     Task to execute when the property changes
     */
    public static void subscribe(SubscriberValidator validator, Object bean,
                                 String propertyPath, final Runnable runnable) {
        String runnableDescription = String.format("Runs %s", runnable);
        subscribe(new PropertyChangeSubscriber(runnableDescription, validator,
                bean, propertyPath, new PropertyChangeListener() {
            @Override
            public void onChange(PropertyChangeMessage message) {
                runnable.run();
            }
        }));
    }

    public static void subscribe(SubscriberValidator validator, final Object bean,
                                 String propertyPath, final PropertyChangeListener listener) {
        String runnableDescription = String.format("Runs %s", listener);
        subscribe(new PropertyChangeSubscriber(runnableDescription, validator,
                bean, propertyPath, listener));
    }

    /**
     * Updates a target property from a source property, using an optional
     * adapter.
     *
     * @param <S>                Type of the source property
     * @param <T>                Type of the target property
     * @param sourceBean         Source bean
     * @param sourcePropertyPath Path to the source property
     * @param targetBean         Target bean
     * @param targetPropertyPath Path to the target property
     * @param adapter            Adapter between the source and the target property, or
     *                           <code>null</code> if no adaptation is needed.
     */
    public static <S, T> void update(Object sourceBean,
                                     String sourcePropertyPath, Object targetBean,
                                     String targetPropertyPath, Adapter<S, T> adapter) {
        Object sourceProperty = Utils.getProperty(sourceBean,
                sourcePropertyPath);
        if (adapter != null) {
            @SuppressWarnings("unchecked")
            S s = (S) sourceProperty;
            sourceProperty = adapter.convertSubjectToTarget(s);
        }
        Utils.setProperty(targetBean, targetPropertyPath, sourceProperty);
    }

    private PropertyChangeSupport() {
    }
}