package net.sf.doolin.bus.support;

import net.sf.doolin.bus.Subscriber;
import net.sf.doolin.bus.SubscriberExecution;
import net.sf.doolin.bus.SubscriberTrigger;
import net.sf.doolin.bus.SubscriberValidator;
import org.apache.commons.lang.Validate;

/**
 * This {@link Subscriber} delegates its methods to a
 * {@link SubscriberValidator}, a {@link SubscriberTrigger} and a
 * {@link SubscriberExecution}.
 *
 * @param <M> Type of message that is handled by this subscriber.
 * @author Damien Coraboeuf
 */
public class SubscriberAdapter<M> implements Subscriber<M> {

    private final SubscriberValidator validator;
    private final SubscriberTrigger<M> trigger;
    private final SubscriberExecution<M> execution;

    /**
     * Constructor.
     *
     * @param validator The validator to use
     * @param trigger   The trigger to use
     * @param execution The execution to use
     */
    public SubscriberAdapter(SubscriberValidator validator,
                             SubscriberTrigger<M> trigger, SubscriberExecution<M> execution) {
        Validate.notNull(validator, "Validator must not be null");
        Validate.notNull(trigger, "Trigger must not be null");
        Validate.notNull(execution, "Execution must not be null");
        this.execution = execution;
        this.trigger = trigger;
        this.validator = validator;
    }

    @Override
    public Class<M> getMessageClass() {
        return trigger.getMessageClass();
    }

    @Override
    public String getExecutionDescription() {
        return this.execution.getExecutionDescription();
    }

    @Override
    public String getTriggerDescription() {
        return this.trigger.getTriggerDescription();
    }

    @Override
    public String getValidatorDescription() {
        return this.validator.getValidatorDescription();
    }

    @Override
    public boolean isValid() {
        return this.validator.isValid();
    }

    @Override
    public void receive(M message) {
        this.execution.receive(message);
    }

    @Override
    public boolean accept(M message) {
        return this.trigger.accept(message);
    }
}
